#!/usr/bin/env python
from ccrenew import __version__
from setuptools import setup

setup(
    name='ccrenew',
    version=__version__,
    author='CCR Perf Eng Team',
    description='Utilities for running Cypress Creek Renewables Performance Engineering workflows',
    url='https://cypresscreekrenew.sharepoint.com/:u:/s/AM-Performance/EdhVQHbS4n9Jm022irFSFiABmo6gcEW3KTmV0qEkfHW84Q?e=zxokuU',
    python_requires='>=3.8'
)